
Number variableByteSubclass: #Float
  classVariableNames: ''
  poolDictionaries: '' !

! Float class methods !

floatError
        "Query the floating point coprocessor as
         to the type of exception and report it."
    | status |
    status := self status.
    (status bitAnd: 8) ~= 0
        ifTrue: [^self error: 'Float overflow exception'].
    (status bitAnd: 16r10) ~= 0
        ifTrue: [^self error: 'Float underflow exception'].
    (status bitAnd: 4) ~= 0
        ifTrue: [^self error: 'Float divide by zero exception'].
    (status bitAnd: 2) ~= 0
        ifTrue: [^self error: 'Float denormalized operand'].
    (status bitAnd: 1) ~= 0
        ifTrue: [^self error: 'Float invalid operation'].
    (status bitAnd: 16r80) ~= 0
        ifTrue: [^self error: 'Math coprocessor missing'].
    ^self error: 'Float undefined exception'!

fromInteger: anInteger
        "Answer a floating point representation
         of the argument anInteger."
    <primitive: 40>
    anInteger isSwapped
        ifTrue: [^self fromInteger: anInteger].
    ^self floatError!

pi
        "Answer the floating point
         representation of pi."
    ^1 arcTan * 4!

status
        "Answer the status of the floating point
         coprocessor as a small integer (refer to
         coprocessor status word definition)."
    <primitive: 86>! !


! Float methods !

* aNumber
        "Answer the result of multiplying
         the receiver by aNumber."
    <primitive: 49>
    aNumber isSwapped
        ifTrue: [^self * aNumber].
    aNumber class == ComplexNumber
        ifTrue: [^aNumber * self].
    aNumber class == Float
        ifTrue: [^self class floatError]
        ifFalse: [^self * aNumber asFloat]!

+ aNumber
        "Answer sum of the receiver and aNumber."
    <primitive: 41>
    aNumber isSwapped
        ifTrue: [^self + aNumber].
    aNumber class == ComplexNumber
        ifTrue:[^aNumber + self].
    aNumber class == Float
        ifTrue: [^self class floatError]
        ifFalse: [^self + aNumber asFloat]!

- aNumber
        "Answer the difference between
         the receiver and aNumber."
    <primitive: 42>
    aNumber isSwapped
        ifTrue: [^self - aNumber].
    aNumber class == ComplexNumber
        ifTrue:[^self asComplex - aNumber].
    aNumber class == Float
        ifTrue: [^self class floatError]
        ifFalse: [^self - aNumber asFloat]!

/ aNumber
        "Answer the result of dividing
         the receiver by aNumber."
    <primitive: 50>
    aNumber isSwapped
        ifTrue: [^self / aNumber].
    aNumber class == ComplexNumber
        ifTrue: [^self asComplex / aNumber].
    aNumber class == Float
        ifTrue: [^self class floatError]
        ifFalse: [^self / aNumber asFloat]!

// aNumber
        "Answer the integer quotient after dividing
         the receiver by aNumber with truncation
         towards negative infinity."
    | answer quotient |
    quotient := self / aNumber.
    (answer := quotient truncated) < 0
        ifTrue: [
            answer = quotient
                ifFalse: [^answer - 1]].
    ^answer!

< aNumber
        "Answer true if the receiver is less
         than aNumber, else answer false."
    <primitive: 43>
    aNumber isSwapped
        ifTrue: [^self < aNumber].
    aNumber class == Float
        ifTrue: [^self class floatError]
        ifFalse: [^self < aNumber asFloat]!

<= aNumber
        "Answer true if the receiver is less than
         or equal to aNumber, else answer false."
    ^(aNumber asFloat < self) not!

= aNumber
        "Answer true if the receiver is equal
         to aNumber, else answer false."
    <primitive: 47>
    aNumber isSwapped
        ifTrue: [^self = aNumber].
    aNumber class == Float
        ifTrue: [^self class floatError]
        ifFalse: [
            (aNumber isKindOf: Number)
                ifTrue: [^self = aNumber asFloat]
                ifFalse: [^false]]!

> aNumber
        "Answer true if the receiver is greater
         than aNumber, else answer false."
    ^aNumber asFloat < self!

>= aNumber
        "Answer true if the receiver is greater than
         or equal to aNumber, else answer false."
    ^(self < aNumber) not!

\\ aNumber
        "Answer the integer remainder after dividing
         the receiver by aNumber with truncation
         towards negative infinity."
    ^(self - (self // aNumber * aNumber))
        truncated!

arcTan
        "Answer the arc-tangent, an angle in
         radians, of the receiver."
    <primitive: 85>
    ^self class floatError!

asFloat
        "Answer the receiver as a floating
         point number."
    ^self!

cos
        "Answer the cosine of the receiver.  The
         receiver is an angle measured in radians"
    <primitive: 30>
    ^self class floatError!

deepCopy
        "Answer the receiver."
    ^self!

degreesToRadians
        "Answer the number of radians the
         receiver represents in degrees."
    ^self * Float pi / 180!

dumpOn: aStream for: aCloner
        "Dump the receiver as a series of
         bytes to aStream for aCloner."
    | size index |
    size := self basicSize.
    aStream nextTwoBytesPut: (aCloner cloneOopFor: self class).
    aStream nextTwoBytesPut: size + 2.
    index := 1.
    [ index <= size ]
        whileTrue: [
            aStream nextBytePut: (self at: index).
            index := index + 1]!

exp
        "Answer the exponential of the receiver."
    <primitive: 45>
    ^self class floatError!

exponent
        "Answer the floating point number whose value
         is the exponent part of the floating
         point representation of the receiver."
    <primitive: 53>
    ^self class floatError!

hash
        "Answer the integer hash value for the receiver."
    | answer index |
    answer := 0.
    index := 8.
    [index <= 4]
        whileFalse: [
            answer := answer + ((self at: index) hash).
            index := index - 1].
    ^answer!

ln
        "Answer the natural log of the receiver."
    <primitive: 46>
    ^self class floatError!

negated
        "Answer the receiver subtracted from zero."
    ^0 asFloat - self!

printOn: aStream
        "Answer the receiver.  Append the ASCII
         representation (maximum of 8 digits) of
         the receiver to aStream."
    | digitCount integer number rounder scale n |
    (number := self) < 0
        ifTrue: [
            aStream nextPut: $-.
            number := self negated].
    rounder := 1/20000000.
    number = 0
        ifTrue: [^aStream nextPutAll: '0.0'].
    scale := 0.
    [number >= 10]
        whileTrue: [
            number := number / 10.
            scale := scale + 1].
    [number < 1]
        whileTrue: [
            number := number * 10.
            scale := scale - 1].
    number := number + rounder.
    number >= 10
        ifTrue: [
            number := number /10.
            scale := scale + 1].
    (scale > 7 or: [
    scale < -7 or: [
    scale < 0 and: [
    n := number * (10 raisedToInteger: 7 + scale).
    ((n - n truncated) * (10 raisedToInteger:
        scale negated)) truncated ~= 0]]])
        ifFalse: [
            number := number * (10 raisedToInteger: scale).
            rounder := rounder * (10 raisedToInteger: scale).
            scale := 0].
    digitCount := aStream position.
    integer := number truncated.
    number := number - integer.
    integer printOn: aStream.
    aStream nextPut: $..
    rounder := rounder + rounder.
    digitCount := 9 - (aStream position - digitCount).
    [number := number * 10.
    integer := number truncated.
    number := number - integer.
    aStream nextPut: (Character digitValue: integer).
    (digitCount := digitCount - 1) > 0 and:
    [(rounder := rounder * 10) <= number]]
        whileTrue: [].
    scale = 0
        ifFalse: [
            aStream nextPut: $e.
            scale printOn: aStream]!

radiansToDegrees
        "Answer the number of degrees the
         receiver represents in radians."
    ^self * 180 / Float pi!

reciprocal
        "Answer one divided by the receiver."
    ^1 asFloat / self!

shallowCopy
        "Answer the receiver."
    ^self!

significand
        "Answer the floating point number whose value
         is the significand part of the floating
         point representation of the receiver."
    <primitive: 52>
    ^self class floatError!

sin
        "Answer the sine of the receiver.  The
         receiver is an angle measured in radians"
    <primitive: 48>
    ^self class floatError!

sqrt
        "Answer the square root of the receiver."
    <primitive: 44>
    ^self class floatError!

tan
        "Answer the tangent of the receiver.  The
         receiver is an angle measured in radians"
    <primitive: 83>
    ^self class floatError!

timesTwoPower: anInteger
        "Answer 2 to the exponent anInteger
         multiplied by the receiver."
    <primitive: 54>
    ^self class floatError!

truncate
        "Private - Answer the receiver as a kind of
         Integer truncating the fraction part."
    | answer float scale partialInt |
    scale := self exponent truncated.
    float := self timesTwoPower: scale \\ 13 - scale.
    answer := float truncated.
    float := float - answer asFloat.
    scale // 13 timesRepeat: [
        float := float timesTwoPower: 13.
        partialInt := float truncated.
        answer := answer * 8192 + partialInt.
        float := (float - partialInt asFloat)].
    ^answer!

truncated
        "Answer the receiver as a kind of
         Integer truncating the fraction part."
    <primitive: 51>
    ^self truncate! !
