
OldExternalStream subclass: #OldFileStream
  instanceVariableNames: 
    'file pageStart writtenOn lastByte lineDelimiter '
  classVariableNames: ''
  poolDictionaries: 
    'CharacterConstants ' !

! OldFileStream class methods ! !


! OldFileStream methods !

atEnd
        "Answer true if the receiver is positioned at the
         end (beyond the last object), else answer false."
    (position < readLimit or:
        [(pageStart + position) <= self updateLastByte])
            ifTrue: [^false]
            ifFalse:[^true]!

close
        "Close the file associated with the receiver stream
         after writing all the data to the file."
    self readLimit.
    self writePage.
    file close!

copyFrom: first to: last
        "Answer a String containing the characters of the
         receiver stream from positions first to last."
    | string savePosition index putIndex number |
    string := collection species new: last - first + 1.
    savePosition := self position.
    putIndex := 1.
    index := first.
    [index <= last]
        whileTrue: [
            self position: index - 1.
            number := (last min: pageStart + readLimit - 1)
                - index + 1.
            string
                replaceFrom: putIndex
                to: putIndex + number - 1
                with: collection
                startingAt: position + 1.
            putIndex := putIndex + number.
            index := index + number].
    self position: savePosition.
    ^string!

cr
        "Write the line terminating character
         (carriage-return line-feed pair or
         line-feed) to the receiver stream."
    self lineDelimiter == Cr
        ifTrue: [self nextPut: Cr].
    self nextPut: Lf!

file
        "Answer the file over which the receiver is
         streaming."
    ^file!

flush
        "Guarantee that any writes to the receiver stream
         are physically recorded on disk."
    self
        setToEnd;
        readLimit;
        writePage.
    file flush!

lineDelimiter
        "Answer the line delimiter character for the
         receiver file stream, either carriage-return
         or line-feed."
    ^lineDelimiter!

lineDelimiter: aCharacter
        "Change the line delimiter character to aCharacter."
    lineDelimiter := aCharacter!

next
        "Answer the next character accessible by the receiver
         and advance the stream position.  Report an error
         if the receiver stream is positioned at end."
    <primitive: 65>
    self position: self position.
    self atEnd
        ifTrue: [
            ^self error: 'read beyond end of stream']
        ifFalse: [
            (position := position + 1) > readLimit
                ifTrue: [self error: 'position outside of stream'].
            ^collection at: position]!

nextLine
        "Answer a String consisting of the characters of
         the receiver up to the next line delimiter."
    | line |
    line := self upTo: lineDelimiter.
    lineDelimiter == Cr
        ifTrue: [self peekFor: Lf].
    ^line!

nextPut: aCharacter
        "Write aCharacter to the receiver stream."
    writtenOn := true.
    ^self primitiveNextPut: aCharacter!

nextPutAll: aCollection
        "Write each of the chararacters in
         aCollection to the receiver stream."
    | currentPosition stopIndex |
    stopIndex := position + aCollection size.
    stopIndex <= writeLimit
        ifTrue: [
            collection
                replaceFrom: position + 1
                to: stopIndex
                with: aCollection
                startingAt: 1.
            position := stopIndex.
            writtenOn := true.
            ^aCollection].
    ^aCollection do: [ :element |
        self nextPut: element]!

pageSize
    ^file pageSize!

pathName
        "Answer the complete pathname of the file
         over which the receiver is streaming."
    | path |
    path := file directory pathName.
    path = '\'
        ifTrue: [path := ''].
    ^(String with: file directory drive), ':',
        path, '\', file name!

position
        "Answer the current receiver stream position."
    ^pageStart + position - 1!

position: anInteger
        "Set the receiver stream position to anInteger."
    self readLimit.
    ((anInteger + 1) < pageStart
        or: [(pageStart + File pageSize) <= (anInteger + 1)])
            ifTrue: [
                self writePage.
                pageStart := ((anInteger // File pageSize)
                    * File pageSize) + 1.
                readLimit := file
                    readBuffer: collection
                    atPosition: pageStart].
    position := anInteger - pageStart + 1.
    self updateLastByte!

primitiveNextPut: aCharacter
        "Private - Write aCharacter to the receiver stream.
         Answer aCharacter."
    <primitive: 66>
    self position: self position.
    position := position + 1.
    writtenOn := true.
    collection at: position put: aCharacter.
    ^aCharacter!

readLimit
        "Private - Answer the size of the receiver file."
    readLimit := readLimit max: position.
    ^self updateLastByte!

setLimits
        "Private - Initialize the stream size and position."
    file := collection.
    collection := String new: File pageSize.
    writeLimit := collection size.
    readLimit := file
        readBuffer: collection
        atPosition: 1.
    position := 0.
    pageStart := 1.
    writtenOn := false.
    lastByte := file size.
    lineDelimiter := Cr.
    2 to: readLimit do: [ :i |
        (collection at: i) == Lf
            ifTrue: [
                (collection at: i - 1) == Cr
                    ifFalse: [lineDelimiter := Lf].
                ^self]]!

updateLastByte
        "Private - Answer the position of the last byte in
         the file, and update the instance variable."
    ^lastByte := lastByte max: position + pageStart - 1!

writePage
        "Private - Write the current page
         to the file, if it has changed."
    writtenOn
        ifTrue: [
            file
                writeBuffer: collection
                ofSize: readLimit
                atPosition: pageStart.
            writtenOn := false]! !
