
Object subclass: #StringModel
  instanceVariableNames: 
    'lines extent lastChild frame topCorner charScanner '
  classVariableNames: ''
  poolDictionaries: 
    'CharacterConstants ' !

!StringModel class methods !

for: aString
        "Answer a new instance of class StringModel
         containing an OrderedCollection of strings
         created from aString by separating it at
         the line-feed characters."
    ^self new string: aString.! !


!StringModel methods !

adjustPoint: aPoint
        "Private - Answer a Point representing
         aPoint adjusted to a valid text position."
    | x y |
    (aPoint x > 0 and: [aPoint y > 0])
        ifTrue: [^aPoint].
    x := aPoint x.  y := aPoint y.
    x < 1 ifTrue: [y := y - 1].
    y < 1 ifTrue: [^1 @ 1].
    x < 1 ifTrue: [
        y <= lines size
            ifTrue: [x := (lines at: y) size + 1]
            ifFalse:[x := 1]].
    ^x @ y!

appendChar: aCharacter
        "Append aCharacter to the end of the last
         line and inform the text pane to update."
    | origin |
    origin := (extent x + 1) @ extent y.
    aCharacter == Lf
        ifTrue: [lines add: String new]
        ifFalse:[
    aCharacter == Tab
        ifTrue: [lines last become: lines last, (self tabStringAt: origin x)]
    ifFalse: [
        lines last become: (lines last copyWith: aCharacter)]].
    extent := self extent.
    lastChild update: (
        TextSelection origin: origin corner: extent)!

appendText: aString
        "Append aString to the end of the last line
         and inform the text pane to update."
    | origin newLine |
    origin := (extent x + 1) @ extent y.
    newLine := lines last.
    newLine become: newLine, aString.
    extent x: newLine size.
    lastChild update: (
        TextSelection origin: origin corner: extent)!

backspacePoint: aPoint
        "Answer a Point describing the position
         of the character preceeding aPoint."
    ^self adjustPoint: aPoint - (1 @ 0)!

broadcastChangesIn: aTextSelection
    upTo: aPoint
    withExcess: excess

        "Private - Inform the text pane to update the changed
         area (a Rectangle) whose corner is the maximum
         of the aTextSelection corner and the sum of aPoint
         and excess."
    | origin corner newExtent |
    origin := aTextSelection origin.
    corner := aTextSelection corner.
    corner = aPoint
        ifFalse: [
            (aPoint isBefore: corner)
                ifTrue: [
                    corner := corner x + excess @ corner y]
                ifFalse:[
                    corner := aPoint x + excess @
                        aPoint y]].
    (newExtent := self extent) = extent
        ifFalse: [
            (newExtent isBefore: extent)
                ifTrue: [corner := extent]
                ifFalse:[corner := newExtent].
            extent := newExtent].
    lastChild update: (
        TextSelection origin: origin corner: corner)!

convertToString: aCollection
        "Private - Answer a String containing the
         concatenation of aCollection of lines."
    | tally index result line i |
    tally := 0.
    1 to: aCollection size do: [ :index |
        line := aCollection at: index.
        (line isEmpty or: [line last ~= $ ])
            ifFalse: [
                i := line size - 1.
                [i > 0 and: [(line at: i) == $ ]]
                    whileTrue: [i := i - 1].
                (i > 0 and: [(line at: i) == $$])
                    ifTrue: [i := i + 1].
                line := line copyFrom: 1 to: i.
                aCollection at: index put: line].
        tally := tally + line size + 1].
    result := String new: tally.
    index := 1.
    aCollection do: [ :line |
        result
            replaceFrom: index
            to: index + line size - 1
            with: line.
        index := index + line size + 1.
        result at: index - 1 put: Lf].
    ^result!

delete: selection
        "Delete the part of the text contained in
         the selection.  Answer a Point describing the
         character position before the deletion."
    | line answer excess |
    answer := selection origin.
    (answer == selection corner
        and: [(lines at: answer y) size >= answer x])
            ifTrue: [^self deleteCharIn: selection].
    line := self rightPartAfter: selection corner.
    excess := line size.
    line := (self leftPartBefore: answer), line.
    answer := answer - (1@0).
    self
        replaceLinesIn: selection with: line;
        broadcastChangesIn: selection
            upTo: answer
            withExcess: excess.
    ^answer!

deleteCharIn: selection
        "Delete the first character in the selection.
         Answer a Point describing the character
         position before the deletion."
    | line answer |
    answer := selection origin.
    line := lines at: answer y.
    lines at: answer y put:
        (line
            copyReplaceFrom: answer x
            to: answer x
            with: '').
    answer := answer - (1@0).
    self
        broadcastChangesIn: selection
        upTo: answer
        withExcess: line size - answer x.
    ^answer!

display: aRectangle
        "Display the text contained in aRectangle."
    | sourceY dest lastY sourceX minY fontSize |
    fontSize := charScanner font charSize.
    sourceX := aRectangle origin x.
    sourceY := aRectangle origin y.
    lastY := aRectangle corner y.
    dest := (self formCoordinate: aRectangle origin).
    lines size >= sourceY ifTrue: [
        (sourceY = lastY) ifTrue: [ "single line"
            charScanner
                display: (lines at: sourceY)
                from: sourceX
                to: aRectangle corner x
                at: dest.
            ^self].
        "display first line for multi lines"
            charScanner
                display: (lines at: sourceY)
                from: sourceX
                at: dest].
    minY := lastY min: lines size.
    sourceY + 1 to: minY do: [ :i |
        dest := 0 @ (dest y + fontSize y).
        charScanner
            display: (lines at: i)
            from: topCorner x
            at: dest].
    minY < lastY
        ifTrue: [
            minY := minY max: sourceY - 1. "for
                case sourceY > minY"
            charScanner
                blankRestFrom:
                (minY - topCorner y + 1
                * fontSize y)]!

displayAll
        "Display the text contained in the pane."
    | destY firstLine lastLine fontHeight |
    firstLine := topCorner y.
    fontHeight := charScanner font height.
    lastLine := frame height //
                fontHeight +
                firstLine.
    lastLine := lastLine min: lines size.
    destY := 0 - fontHeight.
    firstLine to: lastLine do: [:i |
        charScanner
            display: (lines at: i)
            from: topCorner x
            at: (0 @ (destY :=
                i - firstLine *
                fontHeight))].
    (destY + fontHeight <
    frame height) ifTrue: [
        charScanner blankRestFrom:
            (destY + fontHeight)]!

extendTo: aPoint
        "Private - If needed, extend lines
         to aPoint by adding empty strings."
    | index |
    index := aPoint y.
    [index > lines size]
        whileTrue: [lines add: (String new: 0)]!

extent
        "Answer a Point describing the position
         of the last charcter in the text."
    lines size <= 0
        ifTrue: [^0 @ 1]
        ifFalse:[^lines last size @ (lines size)]!

fileInFrom: aStream
        "Replace all the receiver lines
         with the contents of aStream."
    lines := OrderedCollection new.
    [aStream atEnd]
        whileFalse: [lines add: aStream nextLine].
    extent := self extent!

fileOutOn: aStream
        "Write the receiver contents to aStream
         appending a line delimiter to each line."
    lines do: [ :aString | aStream nextPutAll: aString; cr]!

formCoordinate: aPoint
        "Answer a Point, the argument converted
         to a position in the pane as opposed to
         a position in the text."
    ^aPoint - topCorner *
        charScanner font charSize!

frame: aRectangle
        "Change the frame to aRectangle."
    frame := aRectangle!

leftPartBefore: aPoint
        "Private - Answer the substring prior
         to aPoint on that point's line."
    | line position |
    position := aPoint x - 1.
    aPoint y > lines size
        ifTrue: [
            ^(String new: position) atAllPut: Space].
    line := lines at: aPoint y.
    position <= line size
        ifTrue: [
            ^line copyFrom: 1 to: (0 max: position)]
        ifFalse: [
            ^line,
                ((String new: position - line size)
                    atAllPut: Space)]!

lineAt: anInteger
        "Answer the String in the
         line indexed by anInteger."
    (lines size >= anInteger and:
        [anInteger > 0])
        ifTrue: [^lines at: anInteger]
        ifFalse: [^'']!

linesIn: aTextSelection
        "Answer an OrderedCollection of
         the lines contained in aTextSelection."
    | someLines origin corner |
    origin := aTextSelection origin.
    corner := aTextSelection corner.
    self extendTo: corner.
    someLines := OrderedCollection new.
    someLines add:
        (self rightPartAfter: origin - (1 @ 0)).
    (corner y - origin y) > 1
        ifTrue: [
            someLines addAll:
                (lines
                    copyFrom: origin y + 1
                    to: corner y - 1)].
    someLines add:
        (self leftPartBefore: corner + (1 @ 0)).
    ^someLines!

replace: aTextSelection withChar: aCharacter
        "Replace the text in aTextSelection with aCharacter.
         Answer a Point describing the position of the new
         character.  Inform the text pane of the change."
    | origin corner line excess |
    aCharacter == Lf
        ifTrue: [^self replaceWithLf: aTextSelection].
    aCharacter == Tab
        ifTrue: [^self replaceWithTab: aTextSelection].
    origin := aTextSelection origin.
    corner := aTextSelection corner.
    self extendTo: corner.
    (corner isBefore: origin)
        ifTrue: [
            excess := self
                replaceGapBefore: origin
                withChar: aCharacter]
        ifFalse: [
            line := self rightPartAfter: corner.
            excess := line size.
            line := (self leftPartBefore: origin)
                 , (String with: aCharacter)
                 , line.
            self replaceLinesIn: aTextSelection with: line].
    self
        broadcastChangesIn: aTextSelection
        upTo: origin
        withExcess: excess.
    ^origin!

replace: aTextSelection withText: aString
        "Replace the text in aTextSelection with aString.
         Answer a Point describing the position of the last
         replacement character.  Inform the text pane
         of the change."
    | newLines line endOfReplacement stream noneYet excess |
    self extendTo: aTextSelection corner.
    (aString isNil or: [aString isEmpty])
        ifTrue: [^self delete: aTextSelection].
    newLines := OrderedCollection new.
    noneYet := true.
    stream := ReadStream on: aString.
    [stream atEnd]
        whileFalse: [
            noneYet
                ifTrue: [
                    noneYet := false.
                    newLines add:
                        (self leftPartBefore: aTextSelection origin),
                            (stream upTo: Lf)]
                ifFalse:[newLines add: (stream upTo: Lf)]].
    noneYet
        ifTrue: [
            newLines add:
                (self leftPartBefore: aTextSelection origin)].
    endOfReplacement := (newLines last size) @
        (aTextSelection top + newLines size - 1).
    line := self rightPartAfter:
            aTextSelection origin + aTextSelection extent - (1@1).
    excess := line size.
    newLines last become: newLines last , line.
    newLines size = 1
        ifTrue: [
            self
                replaceLinesIn: aTextSelection
                with: newLines first ]
        ifFalse: [
            lines := lines
                replaceFrom: aTextSelection top
                to: aTextSelection top + aTextSelection height - 1
                with: newLines].
    self
        broadcastChangesIn: aTextSelection
        upTo: endOfReplacement
        withExcess: excess.
    ^endOfReplacement!

replaceAtColumns: aPoint by: aString startAt: aTextSelection
        "Replace the line contents between the coordinates
         of aPoint with aString in aTextSelection.
         Answer a TextSelection of the new string."
    | row col |
    row := aTextSelection origin y min: lines size + 1.
    row to: lines size do: [:i |
        lines at: i put:
            (self leftPartBefore: aPoint x @ i), aString,
            (self rightPartAfter: aPoint y @ i)].
    col := aPoint x + aString size.
    ^aTextSelection
        origin: col @ lines size
        corner: col - 1 @ lines size!

replaceGapBefore: aPoint withChar: aCharacter
        "Insert aCharacter before aPoint.
         Answer the number of characters
         to the right of aPoint on the insertion line."
    | line position |
    position := aPoint x - 1.
    line := lines at: aPoint y.
    position = line size
        ifTrue: [
            line become: (line, (String with: aCharacter))]
        ifFalse: [position < line size
            ifTrue: [
                line become:
                    (line
                        copyReplaceFrom: position + 1
                        to: position
                        with: (String with: aCharacter))]
            ifFalse: [
                line become:
                    (self leftPartBefore: aPoint + (1 @ 0)).
                line at: line size put: aCharacter]].
    ^line size - aPoint x!

replaceLinesIn: aTextSelection with: aString
        "Replace the selected text with aString."
    | top height |
    top := aTextSelection top.
    ((height := aTextSelection height) = 1)
        ifTrue: [lines at: top put: aString]
        ifFalse: [
            lines := lines
                replaceFrom: top
                to: top + height - 1
                with: (OrderedCollection with: aString)]!

replaceWithLf: aTextSelection
        "Replace the text in aTextSelection with the
         line-feed character.  Answer the position
         of the new line.  Inform the text pane of
         the change."
    | origin corner endOfReplacement excess |
    origin := aTextSelection origin.
    corner := aTextSelection corner.
    self extendTo: corner.
    lines := lines
        replaceFrom: origin y
        to: corner y
        with:
            (OrderedCollection
                with: (self leftPartBefore: origin)
                with: (self rightPartAfter: corner)).
    endOfReplacement := 0 @ (origin y + 1).
    self
        broadcastChangesIn: aTextSelection
        upTo: endOfReplacement
        withExcess: (lines at: origin y + 1) size.
    ^endOfReplacement!

replaceWithTab: aTextSelection
        "Replace the text in aTextSelection with the
         Tab character.  Answer the position of the
         Tab.  Inform the text pane of the change."
    ^self
        replace: aTextSelection
        withText: (self tabStringAt: aTextSelection origin x)!

rightPartAfter: aPoint
        "Private - Answer a String containing the
         characters to the right of aPoint."
    | aString |
    aString := lines at: aPoint y.
    aString size <= aPoint x
        ifTrue: [^String new: 0]
        ifFalse: [
            ^aString
                copyFrom: aPoint x + 1
                to: aString size]!

scanner: aCharacterScanner
        "Set scanner to aCharacterScanner."
    charScanner := aCharacterScanner!

searchBack: aTextSelection for: aPattern
        "Search backward for aPattern starting from
         the end of aTextSelection.  Answer the
         matched selection if found, else answer nil."
    | row col answer |
    row := aTextSelection corner y min: lines size.
    col := aTextSelection corner x + 1.
    row to: 1 by: -1 do: [ :i |
        (answer := aPattern match: (lines at: i) index: col)
            isNil ifFalse: [
                ^aTextSelection
                    origin: answer x @ i
                    corner: answer y @ i].
        col := 1].
    ^nil!

searchFrom: aTextSelection for: aPattern
        "Search for aPattern starting from
         the end of aTextSelection.  Answer the
         matched selection if found, else answer nil."
    | row col answer |
    row := aTextSelection corner y.
    col := aTextSelection corner x + 1.
    row to: lines size do: [ :i |
        (answer := aPattern match: (lines at: i) index: col)
            isNil ifFalse: [
                ^aTextSelection
                    origin: answer x @ i
                    corner: answer y @ i].
        col := 1].
    ^nil!

string
        "Answer a String containing
         the receiver contents."
    ^self convertToString: lines!

string: aString
        "Change the receiver contents to aString
         (lines are separated by line-feeds)."
    extent := 0@0.
    lines := OrderedCollection new.
    self
        replace: (TextSelection origin: 1@1 corner: 0@1)
        withText: aString!

stringIn: aTextSelection
        "Answer a String which concatenates all
         the lines contained in the aTextSelection."
    |origin corner |
    origin := aTextSelection origin.
    corner := aTextSelection corner.
    self extendTo: corner.
    (corner isBefore: origin)
        ifTrue: [^String new: 0]
        ifFalse:[
            origin y = corner y
                ifTrue: [
                    ^(self leftPartBefore: corner + (1@0))
                        copyFrom: origin x
                        to: corner x]
                ifFalse: [
                    ^self convertToString:
                        (self linesIn: aTextSelection)]]!

tabStringAt: anInteger
        "Private - Answer a String of spaces which spans
         from position anInteger to the next tab stop."
    ^(String new: 4 - ((anInteger - 1) \\ 4) )
        atAllPut: Space!

textPane: aTextPane
        "Associates aTextPane to the receiver
         by setting the lastChild to it."
    lastChild := aTextPane!

topCorner: aPoint
        "Change topCorner to aPoint."
    topCorner := aPoint!

totalLength
        "Answer the total length of
         the data held by the receiver."
    ^lines size! !
